
class Measure(object):

    def __init__(self, label):
        self.label = label
        self._db = None
        self._spc = None

class Sum(Measure):

    def increment(self, old_value, new_value):
        return old_value + new_value

    def fetch(self, **point):
        key = self._spc.key(point)
        return self._db.get(key)[self._name]



import errno
import os
from json import loads, dumps, dump, load
from collections import defaultdict
from contextlib import contextmanager
import sqlite3
try:
    from leveldb import LevelDB, WriteBatch
except:
    pass


import space


MAX_CACHE = 10**4


class Backend(object):

    def __init__(self, meta):
        self._read_cache = {}
        self._write_cache = {}
        self.meta = meta

    def get(self, key):
        if key in self._write_cache:
            return self._write_cache[key]

        if key in self._read_cache:
            return self._read_cache[key]

        value = self.db_get(key)
        if value is None:
            value = defaultdict(float)

        if len(self._read_cache) > MAX_CACHE:
            self._read_cache.clear()
        self._read_cache[key] = value
        return value

    def set(self, key, value):
        if len(self._write_cache) > MAX_CACHE:
            self.flush()
        self._write_cache[key] = value

    def flush(self):
        self.flush_write_cache()
        self._write_cache.clear()

    def flush_write_cache(self):
        raise NotImplementedError()

    def db_get(self, key):
        raise NotImplementedError()

    def close(self):
        self.meta.close()
        self.flush()


class LevelDBBackend(Backend):

    def __init__(self, db_path, meta):
        self.ldb = LevelDB(db_path)
        super(LevelDBBackend, self).__init__(meta)

    def db_get(self, key):
        try:
            return loads(self.ldb.Get(key))
        except KeyError:
            return None

    def flush_write_cache(self):
        batch = WriteBatch()
        for key, value in self._write_cache.iteritems():
            batch.Put(key, dumps(value))
        self.ldb.Write(batch)
        #XXX instead of clearing everything, store keys in different
        #dict wrt key length, and then clear dicts with bigger lenght

class SqliteBackend(Backend):

    def __init__(self, db_path, meta):
        db_file = db_path + 'sqlite'
        self.connection = sqlite3.connect(db_file)
        self.cursor = self.connection.cursor()
        self.cursor.execute('CREATE TABLE IF NOT EXISTS data (' \
            'key TEXT,' \
            'value TEXT,' \
            'PRIMARY KEY(key))')
        super(SqliteBackend, self).__init__(meta)

    def db_get(self, key):
        self.cursor.execute('SELECT value FROM data WHERE key = ?', (key,))
        res = self.cursor.fetchone()
        if res is None:
            return None
        return loads(res[0])

    def flush_write_cache(self):
        all_values = (
            (k, dumps(v)) for k, v in self._write_cache.iteritems()
            )
        self.cursor.executemany('INSERT OR REPLACE into data (key, value)'\
                                    ' VALUES (?, ?)', all_values);
        # self.connection.commit()

    def close(self):
        super(SqliteBackend, self).close()
        self.connection.commit()


class MetaData(object):

    def __init__(self, path):
        self.path = path
        if os.path.exists(path):
            with open(path) as fh:
                disk_data = load(fh)
        else:
            disk_data = {}

        self.data = defaultdict(lambda: defaultdict(set))
        for dim, subdict in disk_data.iteritems():
            for key, value in subdict.iteritems():
                self.data[dim][key].update(value)

    def close(self):
        disk_data = {}
        for dim, subdict in self.data.iteritems():
            disk_data[dim] = {}
            for key, value in self.data[dim].iteritems():
                disk_data[dim][key] = list(value)

        with open(self.path, 'w') as fh:
            dump(disk_data, fh)

    def __getitem__(self, name):
        return self.data[name]


def get_db(uri, name, backend):

    data_path = os.path.join(uri, name)
    try:
        os.makedirs(data_path)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise

    meta_path = os.path.join(data_path, 'meta')
    db_path = os.path.join(data_path, 'db')

    meta = MetaData(meta_path)
    if backend == 'sqlite':
        db = SqliteBackend(db_path, meta)
    elif backend ==  'leveldb':
        db = LevelDBBackend(db_path, meta)
    else:
        raise Exception('Backend %s not known' % backend)
    return db


@contextmanager
def connect(uri, backend='leveldb'):
    for name, spc in space.SPACES.iteritems():
        db = get_db(uri, name, backend=backend)
        spc.set_db(db)
    yield
    for name, spc in space.SPACES.iteritems():
        spc._db.close()




from copy import copy
from itertools import product, izip, imap
from collections import namedtuple
from json import dumps

import common
import dimension
import measure

SPACES = {}

class MetaSpace(type):

    def __new__(cls, name, bases, attrs):
        if not '_name' in attrs:
            attrs['_name'] = name

        for b in bases:
            if not type(b) == cls:
                continue
            if hasattr(b, '_dimensions'):
                for name, dim in b._dimensions.iteritems():
                    attrs[name] = copy(dim)

            if hasattr(b, '_measures'):
                for name, msr in b._measures.iteritems():
                    attrs[name] = copy(msr)

        dimensions = {}
        measures = {}
        for k, v in attrs.iteritems():
            # Collect dimensions
            if isinstance(v, dimension.Dimension):
                dimensions[k] = v
                v._name = k

            # Collect measures
            if isinstance(v, measure.Measure):
                measures[k] = v
                v._name = k

        attrs['_dimensions'] = dimensions
        attrs['_measures'] = measures
        spc = super(MetaSpace, cls).__new__(cls, name, bases, attrs)

        for msr in measures.itervalues():
            msr._spc = spc

        if bases:
            SPACES[attrs['_name']] = spc

        return spc


class Space:

    __metaclass__ = MetaSpace
    _db = None

    @classmethod
    def set_db(cls, db):
        cls._db = db
        for dim in cls._dimensions.itervalues():
            dim._db = db
        for msr in cls._measures.itervalues():
            msr._db = db

    @classmethod
    def aggregates(cls, point):
        for name, dim in cls._dimensions.iteritems():
            yield dim.aggregates(point[name])

    @classmethod
    def load(cls, points):
        for point in points:

            for name, dim in cls._dimensions.iteritems():
                dim.store_coordinate(point[name])

            for parent_coords in product(*(cls.aggregates(point))):
                cls.increment(parent_coords, point)

    @classmethod
    def increment(cls, coords, point):
        key = cls.serialize(coords)
        values = cls._db.get(key)
        for name, measure in cls._measures.iteritems():
            values[name] = measure.increment(values[name], point[name])

        cls._db.set(key, values)

    @classmethod
    def key(cls, point):
        return cls.serialize([
                point.get(name, dim.default) \
                    for name, dim in cls._dimensions.iteritems()
                ])

    @classmethod
    def serialize(cls, point, fill_default=True):
        return dumps(point)

    @classmethod
    def fetch(cls, *measures, **point):
        if len(measures) == 1:
            return cls._measures[measures[0]].fetch(**point)
        return tuple(cls._measures[m].fetch(**point) for m in measures)




#!/usr/bin/env python

#This file is part of Menger. The COPYRIGHT file at the top level of
#this repository contains the full copyright notices and license
#terms.

from distutils.core import setup

setup(name='Menger',
      version='1.1',
      description='Statistics storage',
      author='Bertrand Chenal',
      author_email='bertrandchenal@gmail.com',
      url='https://bitbucket.org/bertrandchenal/menger',
      packages=['menger'],
      requires=['leveldb'],
      license='ISC',
      classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: ISC License (ISCL)',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules',
        ],
      )



from itertools import islice
from json import dumps, loads
from collections import defaultdict
import calendar

class Dimension(object):

    def __init__(self, label):
        self.label = label
        self._db = None

    def serialize(self, coord):
        return dumps(coord)


class Tree(Dimension):

    default = []

    def aggregates(self, coord):
        for i in xrange(len(coord) + 1):
            yield coord[:i]

    def drill(self, coord=[]):
        for res in self._db.meta[self._name][self.serialize(coord)]:
            yield coord + [res]

    def store_coordinate(self, coord):
        for pos, item in enumerate(coord):
            prefix = self.serialize(coord[:pos])
            self._db.meta[self._name][prefix].add(item)

class Flat(Tree):

    default = None

    def aggregates(cls, coord):
        yield None
        yield coord

    def drill(self):
        return self._db.meta[self._name][self.serialize(None)]

    def store_coordinate(self, coord):
        self._db.meta[self._name][self.serialize(None)].add(coord)
